package hr.fer.grading;

import static org.assertj.core.api.Assertions.assertThat;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.zip.ZipFile;

import org.gradle.testkit.runner.BuildResult;
import org.gradle.testkit.runner.GradleRunner;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.io.TempDir;

class GradingPluginFunctionalTest {
  @TempDir
  Path projectDir;

  private File getBuildFile() {
    return projectDir.resolve("build.gradle").toFile();
  }

  private File getSettingsFile() {
    return projectDir.resolve("settings.gradle").toFile();
  }

  private void copyLibs() throws IOException {
    File libsDir = projectDir.resolve("libs").toFile();
    libsDir.mkdirs();

    File sourceLibsDir = new File("libs");
    for(File jar: sourceLibsDir.listFiles()) {
      org.gradle.internal.impldep.com.google.common.io.Files.copy(jar, new File(libsDir, jar.getName()));
    }
  }

  @Test
  void canRunGradeTask() throws IOException {
    copyLibs();
    writeString(getSettingsFile(), "");
    createBuildFile();

    // Run the build
    GradleRunner runner = GradleRunner.create();
    runner.forwardOutput();
    runner.withPluginClasspath();
    runner.withArguments("grade"); // You can add "--debug"
    runner.withProjectDir(projectDir.toFile());
    BuildResult result = runner.build();

    // Verify the result
    String output = result.getOutput();
    assertThat(output)
      .contains("--> Reducing 30%")
      .contains("Total percent: 70%");
  }

  @Test
  void canRunGradeTaskWithGermanLocale() throws IOException {
    copyLibs();
    writeString(getSettingsFile(), "");
    createBuildFile();
    writeString(projectDir.resolve("gradle.properties"), """
            org.gradle.jvmargs=-Duser.language=de
            """);

    // Run the build
    GradleRunner runner = GradleRunner.create();
    runner.forwardOutput();
    runner.withPluginClasspath();
    runner.withArguments("grade"); // You can add "--debug"
    runner.withProjectDir(projectDir.toFile());
    BuildResult result = runner.build();

    // Verify the result
    String output = result.getOutput();
    assertThat(output)
      .contains("Total percent: 70%");
  }

  @Test
  void createSubmission_createsZip() throws IOException {
    copyLibs();
    createSettingsFile("exercise1");
    createBuildFile();
    createJavaFile();
    createResources();
    createExerciseHtmlPage();
    createTestJarFile();

    // Run createSubmission
    runGradle("createSubmission");

    // Verify the result
    Path submissionZip = projectDir.resolve("build/submission/sol_exercise1.zip");
    assertThat(submissionZip)
      .exists();

    ZipFile zipFile = new ZipFile(submissionZip.toFile());
    Set<String> fileNamesInZip = zipFile.stream()
      .map(ze -> ze.getName())
      .collect(Collectors.toSet());

    assertThat(fileNamesInZip)
      .contains("build.gradle")
      .contains("settings.gradle")
      .contains("index.html")
      .contains("res/vidljivost.png")
      .contains("libs/tests.jar")
      .contains("src/main/java/hr/fer/SomeTest.java")
      .contains("src/main/resources/some.properties");
  }

  private BuildResult runGradle(String taskName) throws IOException {
    GradleRunner runner = GradleRunner.create();
    runner.withDebug(true);
    runner.forwardOutput();
    runner.withPluginClasspath();
    runner.withArguments(taskName/*, "--stacktrace"*/);
    runner.withProjectDir(projectDir.toFile());
    BuildResult result = runner.build();
    return result;
  }

  private void createTestJarFile() throws IOException {
    Path libs = projectDir.resolve("libs");
    Files.createDirectories(libs);
    writeString(libs.resolve("tests.jar"), """
            jar file
            """);  }

  private void createExerciseHtmlPage() throws IOException {
    writeString(projectDir.resolve("index.html"), """
        <html>
        <body>
        Exercise text
        </body>
        </html>
        """);

    Path res = projectDir.resolve("res");
    Files.createDirectories(res);
    writeString(res.resolve("vidljivost.png"), """
            vidljivost
            """);
  }


  private void createResources() throws IOException {
    Path testResourcesPath = projectDir.resolve("src/main/resources");
    Files.createDirectories(testResourcesPath);
    writeString(testResourcesPath.resolve("some.properties"), """
            prop1=p1
            prop2=p2
            """);
  }

  private void createJavaFile() throws IOException {
    Path packagePath = projectDir.resolve("src/main/java/hr/fer");
    Files.createDirectories(packagePath);
    writeString(packagePath.resolve("SomeTest.java"), """
            package hr.fer;
            import static org.junit.jupiter.api.Assertions.fail;

            import org.junit.jupiter.api.DisplayName;
            import org.junit.jupiter.api.Tag;
            import org.junit.jupiter.api.Test;

            import hr.fer.junit.grading.Grading;

            class SomeTest {

              @Grading(testGroup = "tag1")
              void test1() {
              }

              @Grading(testGroup = "tag1")
              @DisplayName("2nd test")
              void test2() {
                fail("failing 2nd test");
              }

            }
            """);
  }

  private void createBuildFile() throws IOException {
    writeString(getBuildFile(), """
            plugins {
              id 'java'
              id 'hr.fer.grading'
            }

            repositories {
              mavenLocal()
              mavenCentral()
            }

            dependencies {
              testImplementation fileTree(dir: 'libs', include: '*.jar')
            }

            java {
              toolchain {
                  languageVersion.set(JavaLanguageVersion.of(21))
              }
            }

            grading {
              solutionPrefix = "sol"
            }
            """);
  }

  private void createSettingsFile(String projectName) throws IOException {
    writeString(getSettingsFile(), "rootProject.name = '" + projectName + "'\n");
  }

  private void writeString(File file, String string) throws IOException {
    try (Writer writer = new FileWriter(file)) {
      writer.write(string);
    }
  }

  private void writeString(Path file, String string) throws IOException {
    Files.writeString(file, string, StandardCharsets.UTF_8);
  }
}
