package hr.fer.grading;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.SocketException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.output.WriterOutputStream;
import org.gradle.api.Action;
import org.gradle.api.Plugin;
import org.gradle.api.Project;
import org.gradle.api.Task;
import org.gradle.api.artifacts.Configuration;
import org.gradle.api.artifacts.ConfigurationContainer;
import org.gradle.api.file.FileCollection;
import org.gradle.api.logging.LogLevel;
import org.gradle.api.plugins.JavaPluginExtension;
import org.gradle.api.plugins.PluginManager;
import org.gradle.api.tasks.JavaExec;
import org.gradle.api.tasks.SourceSet;
import org.gradle.api.tasks.SourceSetContainer;
import org.gradle.api.tasks.bundling.Zip;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class GradingPlugin implements Plugin<Project> {
  private static final String JUNIT_GRADING = "hr.fer.grading:junit-grading:";
  private static final String LOGGING = "org.apache.logging.log4j:log4j-slf4j-impl:2.17.1";

  private Logger log = LoggerFactory.getLogger(getClass());

  @Override
  public void apply(Project project) {
    GradingExtension extension = initializeExtension(project);
    configureDependencies(project);
    registerTasks(project, extension);
  }

  private GradingExtension initializeExtension(Project project) {
    GradingExtension extension = project.getExtensions().create("grading", GradingExtension.class);

    extension.getSolutionPrefix().set("solution");
    return extension;
  }

  private void registerTasks(Project project, GradingExtension extension) {
    project.getTasks().register("grade", JavaExec.class, task -> {
      task.setGroup("grading");
      task.dependsOn("testClasses");

      initializeJavaPlugin(project);

      task.setClasspath(findTestClasspath(project));
      task.getMainClass().set("hr.fer.grading.GradingRunner");
      task.setArgs(createArgs());
      task.getLogging().captureStandardError(LogLevel.DEBUG);
      StringWriter sw = new StringWriter();
      task.setStandardOutput(new WriterOutputStream(sw, StandardCharsets.UTF_8));
      // Uncomment for remote debugging
      //task.setDebug(true);

      task.doLast(t -> {
        System.out.println(sw.toString());
        Path resultPath = project.getBuildDir().toPath().resolve("gradingResult.txt");
        try {
          if(Files.exists(resultPath))
            Files.delete(resultPath);
          if(!Files.exists(resultPath.getParent()))
            Files.createDirectories(resultPath.getParent());

          Files.writeString(resultPath, sw.toString(), StandardCharsets.UTF_8);
        } catch (IOException e) {
          throw new RuntimeException("Can not create result file.", e);
        }
      });
    });

    project.getTasks().register("createSubmission", Zip.class, task -> {
      task.setGroup("grading");

      task.getArchiveFileName().set(extension.getSolutionPrefix().get() + "_" + project.getName() + ".zip");
      task.getDestinationDirectory().set(project.getBuildDir().toPath().resolve("submission").toFile());

      task.from(project.fileTree(project.getProjectDir())
          .exclude(project.getBuildDir().getName() + "/**")
          .exclude("bin/**")
          .exclude(".gradle/**"));

      task.doFirst(new Action<Task>() {
        @Override
        public void execute(Task zipTask) {
          createSubmissionTelemetryDirectory(project);
        }
      });

      task.doLast(new Action<Task>() {
        @Override
        public void execute(Task zipTask) {
          deleteSubmissionTelemetryDirectory(project);
        }
      });
    });
  }

  private void deleteSubmissionTelemetryDirectory(Project project) {
    try {
      FileUtils.deleteDirectory(new File(project.getProjectDir(), "submissionTelemetry"));
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  private void createSubmissionTelemetryDirectory(Project project) {
    try {
      File telemetryDirectory = new File(project.getProjectDir(), "submissionTelemetry");
      File netInterfaces = new File(telemetryDirectory, "networkInterfaces.txt");
      FileUtils.createParentDirectories(netInterfaces);

      // write to network interfaces file
      try (PrintWriter pw = new PrintWriter(netInterfaces)) {
        printNetworkTelemetry(pw);
      }

      // copy grading file
      File gradingSourceFile = new File(project.getBuildDir(), "gradingResult.txt");
      if(gradingSourceFile.exists()) {
        FileUtils.copyFile(gradingSourceFile,
            new File(telemetryDirectory, "gradingResult.txt"));
      }

    } catch (Exception e) {
      e.printStackTrace();
    }

  }

  private void printNetworkTelemetry(PrintWriter pw) {
    try {
      NetworkInterfacePrinter printer = new NetworkInterfacePrinter(pw);
      printer.print();
    } catch (SocketException se) {
      se.printStackTrace(pw);
    }
  }

  private List<String> createArgs() {
    return List.of("-tg", "all", "-p", "hr");
  }

  private FileCollection findTestClasspath(Project project) {
    JavaPluginExtension javaExtension = project.getExtensions().findByType(JavaPluginExtension.class);

    final SourceSetContainer sourceSets = javaExtension.getSourceSets();

    final SourceSet testSet = sourceSets.findByName("test");
    FileCollection testClassPath = testSet.getRuntimeClasspath();
    if(log.isDebugEnabled()) {
      StringBuilder sb = new StringBuilder();
      sb.append("***Classpath:\n");
      testClassPath.forEach(file -> {
        sb.append(" - ");
        sb.append(file);
        sb.append("\n");
      });
      log.debug(sb.toString());
    }
    return testClassPath;
  }

  private void configureDependencies(Project project) {
    ConfigurationContainer configurations = project.getConfigurations();

    if(log.isDebugEnabled()) {
      log.debug("Configurations:");
      configurations.forEach(c -> log.debug(" - {}", c));
    }

    Configuration testConfiguration = configurations.findByName("testImplementation");
    testConfiguration.getDependencies().add(
        project.getDependencies().create(JUNIT_GRADING + Version.VERSION));

    testConfiguration.getDependencies().add(
        project.getDependencies().create(Version.MAVEN_GROUP + ":" + Version.MAVEN_NAME + ":" + Version.VERSION));
    testConfiguration.getDependencies().add(
        project.getDependencies().create(LOGGING));

    //configurations.findByName("testRuntimeClasspath").resolve();
  }

  private void initializeJavaPlugin(Project project) {
    // apply plugin Java if it has not been applied so far
    PluginManager pluginManager = project.getPluginManager();
    if (!pluginManager.hasPlugin("java"))
      pluginManager.apply("java");
  }
}
